@@ -12,21 +12,26 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import defaultdict, deque
+import atexit
+from collections import deque
+from functools import partial
 import logging
 import os
 import socket
 import sys
 from threading import Event, Lock, Thread
+
+from six import BytesIO
+from six.moves import range
+
 from errno import EALREADY, EINPROGRESS, EWOULDBLOCK, EINVAL, EISCONN, errorcode
+try:
+    from weakref import WeakSet
+except ImportError:
+    from cassandra.util import WeakSet  # noqa
 
 import asyncore
 
-try:
-    from cStringIO import StringIO
-except ImportError:
-    from StringIO import StringIO  # ignore flake8 warning: # NOQA
-
 try:
     import ssl
 except ImportError:
@@ -35,56 +40,79 @@ except ImportError:
 from cassandra import OperationTimedOut
 from cassandra.connection import (Connection, ConnectionShutdown,
                                   ConnectionException, NONBLOCKING)
-from cassandra.decoder import RegisterMessage
+from cassandra.protocol import RegisterMessage
 from cassandra.marshal import int32_unpack
 
 log = logging.getLogger(__name__)
 
-_loop_started = False
-_loop_lock = Lock()
 
-_starting_conns = set()
-_starting_conns_lock = Lock()
+class AsyncoreLoop(object):
 
+    def __init__(self):
+        self._loop_lock = Lock()
+        self._started = False
+        self._shutdown = False
 
-def _run_loop():
-    global _loop_started
-    log.debug("Starting asyncore event loop")
-    with _loop_lock:
-        while True:
-            try:
-                asyncore.loop(timeout=0.001, use_poll=True, count=None)
-            except Exception:
-                log.debug("Asyncore event loop stopped unexepectedly", exc_info=True)
-                break
+        self._conns_lock = Lock()
+        self._conns = WeakSet()
 
-            with _starting_conns_lock:
-                if not _starting_conns:
+    def maybe_start(self):
+        should_start = False
+        did_acquire = False
+        try:
+            did_acquire = self._loop_lock.acquire(False)
+            if did_acquire and not self._started:
+                self._started = True
+                should_start = True
+        finally:
+            if did_acquire:
+                self._loop_lock.release()
+
+        if should_start:
+            thread = Thread(target=self._run_loop, name="cassandra_driver_event_loop")
+            thread.daemon = True
+            thread.start()
+            atexit.register(partial(self._cleanup, thread))
+
+    def _run_loop(self):
+        log.debug("Starting asyncore event loop")
+        with self._loop_lock:
+            while True:
+                try:
+                    asyncore.loop(timeout=0.001, use_poll=True, count=1000)
+                except Exception:
+                    log.debug("Asyncore event loop stopped unexepectedly", exc_info=True)
                     break
 
-        _loop_started = False
-        if log:
-            # this can happen during interpreter shutdown
-            log.debug("Asyncore event loop ended")
+                if self._shutdown:
+                    break
 
+                with self._conns_lock:
+                    if len(self._conns) == 0:
+                        break
 
-def _start_loop():
-    global _loop_started
-    should_start = False
-    did_acquire = False
-    try:
-        did_acquire = _loop_lock.acquire(False)
-        if did_acquire and not _loop_started:
-            _loop_started = True
-            should_start = True
-    finally:
-        if did_acquire:
-            _loop_lock.release()
+            self._started = False
 
-    if should_start:
-        t = Thread(target=_run_loop, name="event_loop")
-        t.daemon = True
-        t.start()
+        log.debug("Asyncore event loop ended")
+
+    def _cleanup(self, thread):
+        self._shutdown = True
+        log.debug("Waiting for event loop thread to join...")
+        thread.join(timeout=1.0)
+        if thread.is_alive():
+            log.warning(
+                "Event loop thread could not be joined, so shutdown may not be clean. "
+                "Please call Cluster.shutdown() to avoid this.")
+
+        log.debug("Event loop thread was joined")
+
+    def connection_created(self, connection):
+        with self._conns_lock:
+            self._conns.add(connection)
+
+    def connection_destroyed(self, connection):
+        with self._conns_lock:
+            self._conns.discard(connection)
 
 
 class AsyncoreConnection(Connection, asyncore.dispatcher):
@@ -93,10 +121,11 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
     module in the Python standard library for its event loop.
     """
 
+    _loop = AsyncoreLoop()
+
     _total_reqd_bytes = 0
     _writable = False
     _readable = False
-    _have_listeners = False
 
     @classmethod
     def factory(cls, *args, **kwargs):
@@ -116,18 +145,17 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
         asyncore.dispatcher.__init__(self)
 
         self.connected_event = Event()
-        self._iobuf = StringIO()
+        self._iobuf = BytesIO()
 
         self._callbacks = {}
-        self._push_watchers = defaultdict(set)
         self.deque = deque()
         self.deque_lock = Lock()
 
-        with _starting_conns_lock:
-            _starting_conns.add(self)
+        self._loop.connection_created(self)
 
         self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
         self.connect((self.host, self.port))
+        self.add_channel()
 
         if self.sockopts:
             for args in self.sockopts:
@@ -136,8 +164,15 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
         self._writable = True
         self._readable = True
 
-        # start the global event loop if needed
-        _start_loop()
+        # start the event loop if needed
+        self._loop.maybe_start()
+
+    def set_socket(self, sock):
+        # Overrides the same method in asyncore. We deliberately
+        # do not call add_channel() in this method so that we can call
+        # it later, after connect() has completed.
+        self.socket = sock
+        self._fileno = sock.fileno()
 
     def create_socket(self, family, type):
         # copied from asyncore, but with the line to set the socket in
@@ -179,8 +214,7 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
         asyncore.dispatcher.close(self)
         log.debug("Closed socket to %s", self.host)
 
-        with _starting_conns_lock:
-            _starting_conns.discard(self)
+        self._loop.connection_destroyed(self)
 
         if not self.is_defunct:
             self.error_all_callbacks(
@@ -189,8 +223,6 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
             self.connected_event.set()
 
     def handle_connect(self):
-        with _starting_conns_lock:
-            _starting_conns.discard(self)
         self._send_options_message()
 
     def handle_error(self):
@@ -266,7 +298,7 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
 
                         # leave leftover in current buffer
                         leftover = self._iobuf.read()
-                        self._iobuf = StringIO()
+                        self._iobuf = BytesIO()
                         self._iobuf.write(leftover)
 
                         self._total_reqd_bytes = 0
@@ -275,14 +307,14 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
                         self._total_reqd_bytes = body_len + 8
                         break
 
-            if not self._callbacks:
+            if not self._callbacks and not self.is_control_connection:
                 self._readable = False
 
     def push(self, data):
         sabs = self.out_buffer_size
         if len(data) > sabs:
             chunks = []
-            for i in xrange(0, len(data), sabs):
+            for i in range(0, len(data), sabs):
                 chunks.append(data[i:i + sabs])
         else:
             chunks = [data]
@@ -296,15 +328,15 @@ class AsyncoreConnection(Connection, asyncore.dispatcher):
         return self._writable
 
     def readable(self):
-        return self._readable or (self._have_listeners and not (self.is_defunct or self.is_closed))
+        return self._readable or (self.is_control_connection and not (self.is_defunct or self.is_closed))
 
-    def register_watcher(self, event_type, callback):
+    def register_watcher(self, event_type, callback, register_timeout=None):
         self._push_watchers[event_type].add(callback)
-        self._have_listeners = True
-        self.wait_for_response(RegisterMessage(event_list=[event_type]))
+        self.wait_for_response(
+            RegisterMessage(event_list=[event_type]), timeout=register_timeout)
 
-    def register_watchers(self, type_callback_dict):
+    def register_watchers(self, type_callback_dict, register_timeout=None):
         for event_type, callback in type_callback_dict.items():
             self._push_watchers[event_type].add(callback)
-        self._have_listeners = True
-        self.wait_for_response(RegisterMessage(event_list=type_callback_dict.keys()))
+        self.wait_for_response(
+            RegisterMessage(event_list=type_callback_dict.keys()), timeout=register_timeout)
